from .db_field import *
from datetime import datetime
from .db_exceptions import ObjectDoesNotExist, MultipleResultFound
import logging as log
from functools import partial
from .db_utils import *
from collections import OrderedDict
from .query import QuerySet, ManyToManySet, OneToManySet
from .rel import Rel, ManyToOneRel, OneToManyRel, ManyToManyRel, many2oneGetter, many2oneSetter
from .platform import coroutine, Return, py3, pyield, pret

TX_SUCCESS = 0
TX_FAIL = 1


class DB_Manager(object):
    select_sql = ''
    insert_sql = ''
    update_sql = ''
    delete_sql = ''

    def __init__(self):
        self._inserts = {}
        self._updates = {}
        self._relationships = {}

    def add_insert(self, attr_name, default=None, auto_now_add=None):
        self._inserts[attr_name] = {'default': default,
                                    'auto_now_add': auto_now_add}

    def add_update(self, attr_name, default=None, auto_now=None):
        self._updates[attr_name] = {'default': default,
                                    'auto_now': auto_now}

    def add_relationship(self, rel_name, relationship):
        if not isinstance(relationship, Rel):
            raise TypeError('add relation_ship fail')
        self._relationships[rel_name] = relationship

    def get_insert_default_val(self, attr_name):
        attr = self._inserts.get(attr_name, None)
        if attr is None:
            return
        default_val = attr.get('default', None)
        if default_val is None:
            auto_now_add = attr.get('auto_now_add', None)
            if auto_now_add:
                default_val = datetime.now()
        return default_val

    def get_update_default_val(self, attr_name):
        attr = self._updates.get(attr_name, None)
        if attr is None:
            return
        default_val = attr.get('default', None)
        if default_val is None:
            auto_now = attr.get('auto_now', None)
            if auto_now:
                default_val = datetime.now()
        return default_val

    def get_relationships(self):
        for rel_name, rel in self._relationships.items():
            yield rel_name, rel


def with_metaclass(meta, *bases):
    class metaclass(meta):
        def __new__(cls, name, this_bases, d):
            return meta(name, bases, d)

    return type.__new__(metaclass, 'temporary_class', (), {})


class ModelMeta(type):
    def __new__(cls, name, bases, attr):
        fields = OrderedDict()
        db_manager = DB_Manager()
        super_new = super(ModelMeta, cls).__new__
        l_name = name.lower()
        db_table = attr.get('db_table', None)
        if db_table is None:
            attr['db_table'] = db_table = l_name
        if len(bases) == 0:
            return super_new(cls, name, bases, attr)

        o2m_rel_set = set()
        m2m_rel_set = set()
        upper_attr = attr.copy()
        db_connection_pool = get_db_connection_pool(name)
        upper_attr['db_connection_pool'] = db_connection_pool
        for k, v in attr.items():
            if isinstance(v, Field):
                db_manager.add_insert(k, v.default, v.auto_now_add)
                db_manager.add_update(k, v.default, v.auto_now)
                upper_attr.__delitem__(k)
                fields[k] = v
            elif isinstance(v, ManyToOneField):
                _fk = v.foreign_key
                if _fk is None:
                    _fk = k + '_id'
                    v.foreign_key = _fk
                _k = '_' + k
                upper_attr[k] = property(partial(many2oneGetter, outter_param=_k),
                                         partial(many2oneSetter, outter_param=_k, foreign_key=_fk))
                rel_cls = v.to
                o2m_rel_set.add((rel_cls, _fk))
                fields[k] = ManyToOneRel(rel_cls, k, _fk)
            elif isinstance(v, ManyToManyField):
                rel_cls = v.to
                mt = v.middle_table
                trf = v.target_rel_field
                srf = v.src_rel_field
                if trf is None:
                    trf = rel_cls.__name__.lower()
                if srf is None:
                    srf = l_name
                db_manager.add_relationship(k, ManyToManyRel(rel_cls, mt, trf, srf))
                rel_attr = k[:k.rfind('_') + 1] + l_name + 's'
                m2m_rel_set.add((rel_cls, rel_attr, mt, srf, trf))
                upper_attr.__delitem__(k)

        select_sql_field_str = "SELECT `%s`.`id`" % db_table
        insert_sql_field_str = ""
        update_sql_field_str = ""
        for field_key, field_value in fields.items():
            if isinstance(field_value, ManyToOneRel):
                f = field_value.foreign_key
            else:
                f = field_key
            select_sql_field_str += ',`%s`.`%s`' % (db_table, f)
            insert_sql_field_str += '`%s`,' % f
            update_sql_field_str += '`%s`.`%s`=' % (db_table, f) + '%s,'
        select_sql = select_sql_field_str + " FROM `%s`" % db_table
        insert_sql = "INSERT INTO `%s`(%s) VALUES(%s)" % (
        db_table, insert_sql_field_str[:-1], '%s,' * (len(fields) - 1) + '%s')
        update_sql = "UPDATE `%s` SET %s WHERE `%s`.`id`=" % (db_table, update_sql_field_str[:-1], db_table) + '%s'

        db_manager.select_sql = select_sql
        db_manager.insert_sql = insert_sql
        db_manager.update_sql = update_sql
        db_manager.delete_sql = 'DELETE FROM `%s` WHERE `%s`.`id`=' % (db_table, db_table) + '%s'

        upper_attr['db_manager'] = db_manager
        upper_attr['fields'] = fields
        new_cls = super_new(cls, name, bases, upper_attr)

        for rel_cls, _fk in o2m_rel_set:
            rel_attr = l_name + '_set'
            rel_cls.db_manager.add_relationship(rel_attr, OneToManyRel(new_cls, _fk, db_connection_pool._kwargs.get(
                'db') != rel_cls.db_connection_pool._kwargs.get('db')))

        for rel_cls, rel_attr, mt, trf, srf in m2m_rel_set:
            rel_cls.db_manager.add_relationship(rel_attr, ManyToManyRel(new_cls, mt, trf, srf))

        return new_cls


class BaseModel(with_metaclass(ModelMeta)):
    db_table = None

    def __init__(self, **kwargs):
        self.id = None
        self.rel_exec_set = set()
        for rel_name, rel in self.db_manager.get_relationships():
            if isinstance(rel, OneToManyRel):
                setattr(self, rel_name, OneToManySet(self, rel))
            elif isinstance(rel, ManyToManyRel):
                setattr(self, rel_name, ManyToManySet(self, rel))
        for k, v in kwargs.items():
            setattr(self, k, v)

    exec("""
@coroutine
def execute_tx(self, sql_str, sql_param, tx):
    try:
        ret = %s tx.execute(sql_str, sql_param)
        if self.id is None:
            self.id = ret._result.insert_id

    except Exception as e:
        log.error("%%s.", e)
        %s tx.rollback()
        ret = TX_FAIL
    else:
        ret = TX_SUCCESS
    %s
    """ % ((pyield,)*2+(pret(),)))

    exec("""
@coroutine
def save(self):
    sql_param = []
    db_manager = self.db_manager
    if self.id is None:
        sql_str = db_manager.insert_sql
        get_default_val = db_manager.get_insert_default_val
    else:
        sql_str = db_manager.update_sql
        get_default_val = db_manager.get_update_default_val
    for fk, fv in self.fields.items():
        if isinstance(fv, ManyToOneRel):
            fk = fv.foreign_key
        v = getattr(self, fk, get_default_val(fk))
        sql_param.append(v)
    if self.id is not None:
        sql_param.append(self.id)

    with (%s self.db_connection_pool.begin()) as tx:
        ret = %s self.execute_tx(sql_str, sql_param, tx)
        if ret is TX_SUCCESS:
            for rel in self.rel_exec_set:
                if isinstance(rel, QuerySet):
                    %s rel.execute(tx)
            self.rel_exec_set.clear()
            %s tx.commit()
    %s
    """ % ((pyield,)*4+(pret(),)))

    exec("""
@coroutine
def delete(self):
    sql_str = self.db_manager.delete_sql
    sql_param = [self.id]
    with (%s self.db_connection_pool.begin()) as tx:
        ret = %s self.execute_tx(sql_str, sql_param, tx)
        if ret is TX_SUCCESS:
            %s tx.commit()
    %s
    """ % ((pyield,)*3+(pret(),)))

    exec("""
@classmethod
@coroutine
def __query(cls, unique, group_fields, order_fields, **where_case):
    select_sql, select_param = get_where_sql(cls,
                                             cls.db_manager.select_sql,
                                             None,
                                             True,
                                             **where_case)

    if group_fields:
        select_sql += 'group by '
        count = len(group_fields)
        for g_case in group_fields:
            select_sql += '%%s ' %% g_case
            count -= 1
            if count > 0:
                select_sql += ','
    if order_fields:
        select_sql += 'order by '
        count = len(order_fields)
        for o_field in order_fields:
            if o_field.startswith('-'):
                select_sql += '%%s desc' %% o_field[1:]
            else:
                select_sql += '%%s asc' %% o_field
            count -= 1
            if count > 0:
                select_sql += ','
    with (%s cls.db_connection_pool.execute(select_sql, select_param)) as cursor:
        datas = cursor.fetchall()
        if unique:
            dl = len(datas)
            if dl == 0:
                raise ObjectDoesNotExist
            elif dl > 1:
                raise MultipleResultFound

        query_list = []
        for data in datas:
            ret = cls()
            ret.id = data[0]
            for idx, field in enumerate(cls.fields.items()):
                fk, fv = field
                val = data[idx + 1]
                if isinstance(fv, ManyToOneRel):
                    rel_obj = None
                    if val is not None:
                        try:
                            rel_obj = %s fv.rel_cls.get(id=val)
                        except ObjectDoesNotExist:
                            pass
                    setattr(ret, fv.rel_name, rel_obj)
                else:
                    setattr(ret, fk, val)
            query_list.append(ret)
    if unique:
        %s
    else:
        %s
    """ % ((pyield,)*2+(pret("query_list[0]"), pret("query_list"))))

    exec("""
@classmethod
@coroutine
def filter(cls, group_fields=None, order_fields=None, **where_case):
    ret = %s cls.__query(False, group_fields, order_fields, **where_case)
    %s

@classmethod
@coroutine
def all(cls, group_fields=None, order_fields=None):
    ret = %s cls.__query(False, group_fields, order_fields)
    %s

@classmethod
@coroutine
def get(cls, **where_case):
    ret = %s cls.__query(True, None, None, **where_case)
    %s
    """ % ((pyield, pret()) * 3))



class Field(object):
    def __init__(self, **kwargs):
        self.default = kwargs.get('default', None)
        self.auto_now_add = kwargs.get('auto_now_add', None)
        self.auto_now = kwargs.get('auto_now', None)


class RelatedField(object):
    one2one = None
    one2many = None
    many2one = None
    many2many = None

    def __init__(self, to, foreign_key=None, middle_table=None, target_rel_field=None, src_rel_field=None):
        self.to = to
        self.foreign_key = foreign_key
        self.middle_table = middle_table
        self.target_rel_field = target_rel_field
        self.src_rel_field = src_rel_field


class OneToOneField(RelatedField):
    one2one = True

    def __init__(self, to, foreign_key=None):
        super(OneToOneField, self).__init__(to=to, foreign_key=foreign_key)


class OneToManyField(RelatedField):
    one2many = True

    def __init__(self, to, foreign_key=None):
        super(OneToManyField, self).__init__(to=to, foreign_key=foreign_key)


class ManyToOneField(RelatedField):
    many2one = True

    def __init__(self, to, foreign_key=None):
        super(ManyToOneField, self).__init__(to=to, foreign_key=foreign_key)


class ManyToManyField(RelatedField):
    many2many = True

    def __init__(self, to, middle_table, target_rel_field=None, src_rel_field=None):
        super(ManyToManyField, self).__init__(to=to, middle_table=middle_table, target_rel_field=target_rel_field, src_rel_field=src_rel_field)


class ForeignKey(ManyToOneField):
    pass


from .db_utils import get_where_sql
from .db_exceptions import ObjectDoesNotExist
from .rel import ManyToOneRel, OneToManyRel, ManyToManyRel
from .platform import coroutine, Return, pyield, pret
import logging as log


class QuerySet(object):
    rel_id = None

    def add(self, val):
        pass

    def remove(self, val):
        pass

    @coroutine
    def filter(self):
        pass

    @coroutine
    def all(self):
        pass

    @coroutine
    def execute(self, tx):
        pass


class OneToManySet(QuerySet):

    def __init__(self, outter, rel):
        if not isinstance(rel, OneToManyRel):
            raise TypeError('invalid relationship')
        self.outter = outter
        self.rel = rel
        self.rel_add_objs = set()
        self.add_rels = set()
        self.remove_rels = set()

    def add(self, obj):
        rel = self.rel
        if not isinstance(obj, rel.rel_cls):
            raise TypeError('add fail')
        oid = obj.id
        if oid is None:
            self.rel_add_objs.add(obj)
        else:
            self.add_rels.add(oid)
        self.outter.rel_exec_set.add(self)

    def remove(self, obj):
        rel = self.rel
        if not isinstance(obj, rel.rel_cls):
            raise TypeError('remove fail')
        oid = obj.id
        if oid is not None:
            self.remove_rels.add(oid)
            self.outter.rel_exec_set.add(self)

    exec("""
@coroutine
def filter(self, **kwargs):
    rid = self.outter.id
    rel = self.rel
    ret = None
    if rid is not None:
        ret = %s eval('rel.rel_cls.filter(%%s=rid, **kwargs)' %% rel.foreign_key)
    %s
    
@coroutine
def all(self):
    ret = %s self.filter()
    %s
    """ % ((pyield, pret()) * 2))

    exec("""
@coroutine
def execute(self, tx):
    rel = self.rel
    rel_cls = rel.rel_cls
    is_across = rel.is_across_db
    if is_across:
        tx = %s rel_cls.db_connection_pool.begin()
    db_manager = rel_cls.db_manager
    rid = self.outter.id
    for obj in self.rel_add_objs:
        sql_param = []
        for fk, fv in rel_cls.fields.items():
            if isinstance(fv, ManyToOneRel):
                fk = fv.foreign_key
                if fk == self.rel.foreign_key:
                    v = rid
                else:
                    v = getattr(obj, fk, None)
            else:
                v = getattr(obj, fk, db_manager.get_insert_default_val(fk))
            sql_param.append(v)
        sql_str = db_manager.insert_sql
        %s obj.execute_tx(sql_str, sql_param, tx)
    self.rel_add_objs.clear()

    add_sql = rel.add_sql
    for oid in self.add_rels:
        try:
            %s tx.execute(add_sql, [rid, oid])
        except Exception as e:
            log.error("%%s.", e)
            %s tx.rollback()
    self.add_rels.clear()

    remove_sql = rel.remove_sql
    for oid in self.remove_rels:
        try:
            %s tx.execute(remove_sql, [oid])
        except Exception as e:
            log.error("%%s.", e)
            %s tx.rollback()
    self.remove_rels.clear()

    if is_across:
        %s tx.commit()
    """ % ((pyield,) * 7))


class ManyToManySet(QuerySet):

    def __init__(self, outter, rel):
        if not isinstance(rel, ManyToManyRel):
            raise TypeError('invalid relationship')
        self.outter = outter
        self.rel = rel
        self.remove_rels = set()
        self.update_rels = set()
        self.rel_add_objs = set()

    def add(self, obj):
        rel = self.rel
        if not isinstance(obj, rel.rel_cls):
            raise TypeError('add fail')
        rid = self.outter.id
        oid = obj.id
        if rid is not None:
            if oid is None:
                self.rel_add_objs.add(obj)
            else:
                self.update_rels.add((rid, oid))
            self.outter.rel_exec_set.add(self)

    def remove(self, obj):
        rel = self.rel
        if not isinstance(obj, rel.rel_cls):
            raise TypeError('remove fail')
        rid = self.outter.id
        oid = obj.id
        if None not in (rid, oid):
            self.remove_rels.add((rid, oid))
            self.outter.rel_exec_set.add(self)

    exec("""
@coroutine
def filter(self, **kwargs):
    rid = self.outter.id
    rel = self.rel
    query_list = None
    if rid is not None:
        rel_cls = rel.rel_cls
        query_list = []
        pool = rel_cls.db_connection_pool
        select_sql, sql_param = get_where_sql(rel_cls, rel.select_sql, [rid], False, **kwargs)
        with (%s pool.execute(select_sql, sql_param)) as cursor:
            for data in cursor.fetchall():
                ret = rel_cls()
                ret.id = data[0]
                for idx, field in enumerate(rel_cls.fields.items()):
                    fk, fv = field
                    val = data[idx + 1]
                    if isinstance(fv, ManyToOneRel):
                        rel_obj = None
                        if val is not None:
                            try:
                                rel_obj = %s fv.rel_cls.get(id=val)
                            except ObjectDoesNotExist:
                                pass
                        setattr(ret, fv.rel_name, rel_obj)
                    else:
                        setattr(ret, fk, val)
                query_list.append(ret)
    %s
    """ % ((pyield,) * 2 + (pret("query_list"),)))

    exec("""
@coroutine
def all(self):
    ret = %s self.filter()
    %s
    """ % (pyield, pret()))

    exec("""
@coroutine
def execute(self, tx):
    rel = self.rel
    rid = self.outter.id
    db_manager = rel.rel_cls.db_manager
    for obj in self.rel_add_objs:
        sql_param = []
        for fk, fv in rel.rel_cls.fields.items():
            if isinstance(fv, ManyToOneRel):
                fk = fv.foreign_key
            v = getattr(obj, fk, db_manager.get_insert_default_val(fk))
            sql_param.append(v)
        sql_str = db_manager.insert_sql
        %s obj.execute_tx(sql_str, sql_param, tx)
        try:
            %s tx.execute(rel.add_sql, [rid, obj.id])
        except Exception as e:
            log.error("%%s.", e)
            %s tx.rollback()
    self.rel_add_objs.clear()

    for rid, oid in self.update_rels:
        is_rel = %s rel.is_related(rid, oid, tx)
        if not is_rel:
            try:
                %s tx.execute(rel.add_sql, [rid, oid])
            except Exception as e:
                log.error("%%s.", e)
                %s tx.rollback()
    self.update_rels.clear()

    for rid, oid in self.remove_rels:
        is_rel = %s rel.is_related(rid, oid, tx)
        if is_rel:
            try:
                %s tx.execute(rel.remove_sql, [rid, oid])
            except Exception as e:
                log.error("%%s.", e)
                %s tx.rollback()
    self.remove_rels.clear()
    """ % ((pyield,) * 9))



import tormysql
from .rel import ManyToOneRel

def init_mysql(**conn):
    global _pool_dict, _database_router
    _pool_dict = {}
    _database_router = {}
    for k, v in conn.items():
        if k != "default":
            tables = v.pop("tables", [])
            _database_router[k] = tables
        _pool_dict[k] = tormysql.helpers.ConnectionPool(**v)


def get_db_connection_pool(cls_name):
    pool = _pool_dict.get("default")
    for k, v in _database_router.items():
        if cls_name.lower() in v:
            pool = _pool_dict.get(k)
            break
    return pool


def get_where_sql(model_cls, sql_prefix, sql_param, plus_where, **where_case):
    db_table = model_cls.db_table
    if sql_prefix is None:
        select_sql = 'SELECT `%s`.* FROM `%s`' % (db_table, db_table)
    else:
        select_sql = sql_prefix
    if sql_param is None:
        select_param = []
    else:
        select_param = sql_param
    if where_case:
        if plus_where:
            select_sql += ' WHERE '
        else:
            select_sql += ' AND '
        count = len(where_case)
        for k, v in where_case.items():
            t = None
            if '__' in k:
                idx = k.rfind('__')
                t = k[idx:]
                k = k[:idx]
            if isinstance(model_cls.fields.get(k, None), ManyToOneRel):
                k = k + '_id'
                if isinstance(v, (list, tuple)):
                    v = [o.id for o in v]
                else:
                    v = v.id
            if t == '__in':
                select_sql += '`%s`.`%s` IN (%s)' % (db_table, k, '%s,' * (len(v) - 1) + '%s')
                select_param.extend(v)
            elif t == '__between':
                select_sql += '`%s`.`%s` > ' % (db_table, k) + "%s AND " + '`%s`.`%s` < ' % (db_table, col) + "%s "
                select_param.extend(v)
            elif t == '__contains':
                select_sql += "`%s`.`%s` LIKE " % (db_table, k) + "\'%%" + str(v) + "%%\' "
            else:
                select_sql += '`%s`.`%s`' % (db_table, k)
                if v is None:
                    select_sql += 'IS NULL '
                else:
                    select_sql += '=%s '
                    select_param.append(v)
            count -= 1
            if count > 0:
                select_sql += 'AND '
    return select_sql, select_param







import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    install_requires=[
        "tormysql ==0.4.0"
    ],
    name="easy_tormysql",
    version="0.2.4",
    author="NightRaven",
    author_email="1453878150@qq.com",
    description="This module provides a twice encapsulation of TorMySql based on Django's code style.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/encodingForBetterWorld/easy_tormysql",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)


import sys

py3 = sys.version_info[0] >= 3
coroutine, Return, pyield = None, None, py3 and "yield from" or "yield"
try:
    import asyncio

    coroutine = asyncio.coroutines.coroutine
except ImportError:
    from tornado.gen import coroutine

    coroutine = coroutine

try:
    import tornado.gen

    Return = tornado.gen.Return
except ImportError:
    pass


def pret(ret="ret"):
    if py3:
        return "return %s" % ret
    else:
        return "raise Return(%s)" % ret



from .platform import coroutine, Return, pyield, pret

class Rel(object):
    pass


class ManyToOneRel(Rel):

    def __init__(self, rel_cls, rel_name, foreign_key):
        self.rel_cls = rel_cls
        self.rel_name = rel_name
        self.foreign_key = foreign_key


class OneToManyRel(Rel):

    def __init__(self, rel_cls, foreign_key, is_across_db):
        self.rel_cls = rel_cls
        self.foreign_key = foreign_key
        self.is_across_db = is_across_db
        db_table = rel_cls.db_table
        prefix = 'UPDATE `%s` SET `%s`.`%s`' % (db_table, db_table, foreign_key)
        suffix = 'WHERE `%s`.`id`=' % db_table + '%s'
        self.add_sql = prefix + '=%s ' + suffix
        self.remove_sql = prefix + '=NULL ' + suffix


class ManyToManyRel(Rel):

    def __init__(self, rel_cls, mt, trf, srf):
        self.rel_cls = rel_cls
        tt = rel_cls.db_table
        select_sql = 'SELECT `%s`.`id`' % tt
        for fk, fv in rel_cls.fields.items():
            if isinstance(fv, ManyToOneRel):
                fk = fv.foreign_key
            select_sql += ',`%s`.`%s`' % (tt, fk)
        select_sql += ' FROM `%s` INNER JOIN `%s` ON (`%s`.`id`=`%s`.`%s_id`) WHERE `%s`.`%s_id`=' \
                   % (tt, mt, tt, mt, trf, mt, srf) + '%s'
        self.select_sql = select_sql

        self.add_sql = 'INSERT INTO `%s`(`%s_id`,`%s_id`) VALUES(' % (mt, srf, trf) + '%s,%s)'

        self.remove_sql = 'DELETE FROM `%s` WHERE `%s_id`=' % (mt, srf) + '%s AND ' + '`%s_id`=' % trf + '%s'

        self.check_sql = 'SELECT COUNT(*) FROM `%s` WHERE `%s_id`=' % (mt, srf) + '%s AND ' + '`%s_id`=' % trf + '%s'

    exec("""
@coroutine
def is_related(self, sid, tid, tx):
    with (%s tx.execute(self.check_sql, [sid, tid])) as cursor:
        count = cursor.fetchall()[0][0]
        ret = count > 0
    %s
    """ % (pyield, pret()))


def many2oneGetter(outter, outter_param):
    return getattr(outter, outter_param, None)


def many2oneSetter(outter, obj, outter_param, foreign_key):
    oid = None
    if obj is not None:
        oid = obj.id
    setattr(outter, foreign_key, oid)
    setattr(outter, outter_param, obj)